#!/usr/bin/env python2.3
"""
Generate a sequence set from given models.

Usage: estimation_with_replications.py [options] <input_data_file.txt>
Options:
    -h          Print this help message.
    -c number   Number of Components
    -e C|M      Type of estimation
                C - clustering
		M - mixture
		K - k-means
    -i R|K|M|L    Initialization method
                R - random initialization
		K - k-means initialization
		L - with labels
		M <model_file.smo> - given model initialization
    -l number   Percentage of labels to be use (number/1000)
    -m file     Model file (only with conjuntion with -i M
    -no         Output statistics for number finding the number of components (AIC,BIC,IBIC)
    -r number   Number of Repetitions
    -s number   Number of states in the random models
    -v number   Use of the viterbi decomposition to find <number> components


Example: > estimation_with_replications.py -c 2 -r 30 -e M -i R tiny.txt

Notes:
 * Requires the ghmm module.
 * If the sequence file already exists, the new sequence set is appended.

"""

import GQLMixture
import GQLCluster
import GQLValidation
import ViterbiDecomposition
import Numeric
import PyclusterInterface
import getopt
from random import *
import MLab
import ghmm

def printTable(table):
        s = ""
	saux = ""
	for j in range(len(contigencyTable[0])):
		s += "%4d" % (j+1)
		saux+="----"
	s += " \n"+saux+"\n"

	for i in range(len(table)):
		s += "%4d |" % (i+1)
		for j in range(len(contigencyTable[i])):
			s += "%4d" % contigencyTable[i,j]
		s +="\n"
	return s

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

class random_models_args:
	def __init__(self):
		self.nr_states = nr_states
		self.copies_each = num_clusters
		self.default_mean =0.0
		self.default_var = 0.2
		self.total_duration = duration
		self.randomize_pars = 0


def set_random_partial_labels_equal(profileClustering,percentageOfLabels,labelsClasses):
   profileClustering.partials = []
   profileClustering.partial_label = {}

   no_classes = len(labelsClasses)
   no_labels = len(profileClustering.profileSet)*percentageOfLabels
   k = int(no_labels/no_classes)
   if k == 0:
       k = 1

   print no_classes, no_labels, percentageOfLabels,k

   for i,c in enumerate(labelsClasses):
       selected = sample(range(max(k,len(c))),k)
       profileClustering.partials.append([])
       for s in selected:
           profileClustering.partials[i].append(c[s])

       for s in selected:
           profileClustering.partial_label[c[s]] = i
   print "labels", profileClustering.partials

def set_random_partial_labels(profileClustering,percentageOfLabels,labelsClasses):
   profileClustering.partials = []
   profileClustering.partial_label = {}

   for i,c in enumerate(labelsClasses):
       print len(c),percentageOfLabels
       k = int(len(c)*percentageOfLabels)
       if k == 0:
            k = 1
       print c,k
       selected = sample(range(len(c)),k)
       profileClustering.partials.append([])
       for s in selected:
           profileClustering.partials[i].append(c[s])

       for s in selected:
           profileClustering.partial_label[c[s]] = i
   print "labels", profileClustering.partials


def replications(profileSet, model,repetitions, initializationMethod,estimationMethod,inputmodel,perlabels):
        currentProfile = None
	profiles = []

	if(initializationMethod in ('L1','L2')):
	    labels = profileSet.seq_classes
    	    labelsClasses = []
            for i in range(max(labels)):
                labelsClasses.append([])
            for i,l in enumerate(labels):
                labelsClasses[l-1].append(i)

	for j in range(repetitions):
  	    try:
		alpha = []


		currentProfile = GQLCluster.ProfileClustering()
		currentProfile.setProfileSet(profileSet)

		if(initializationMethod == 'R'):
		    currentProfile.randomModels(model)
		    currentProfile.estimateFromRandomWeights()
		elif(initializationMethod == 'K'):
		     # estimate from kmeans results
		    currentProfile.randomModels(model)
		    [clusterid, centroids, error, nfound] = PyclusterInterface.kmeans(profileSet,num_clusters)
		    currentProfile.estimateFromRandomWeights(1,assignment=clusterid)
		elif(initializationMethod in ('L1','L2','L3')):
		     # estimate from kmeans results
		    currentProfile.randomModels(model)
		    if(initializationMethod == 'L1'):
		        set_random_partial_labels(currentProfile,perlabels,labelsClasses)
                    elif(initializationMethod == 'L2'):
		        set_random_partial_labels_equal(currentProfile,perlabels,labelsClasses)
                    elif(initializationMethod == 'L3'):
		    	currentProfile.ReadPartialAssignment(perlabels)
                    currentProfile.estimateFromPartialAssignment() # estimate from random assigments
		else:
		    currentProfile.readModels(inputmodel)

		if(initializationMethod in ('L1','L2','L3')):
  		    [currentProfile.ml,
		     currentProfile.alpha,
		     currentProfile.p] =    GQLMixture.estimate_mixture_partials(currentProfile.modelList(),
			    					        profileSet.ghmm_seqs, 10, 0.01, 					 			                        				        currentProfile.fixed_models,
									currentProfile.partial_label)
	            currentProfile.cluster = GQLMixture.decode_mixture(currentProfile.p, num_clusters)
		elif(estimationMethod == 'M'):

  		    [currentProfile.ml,
		     currentProfile.alpha,
		     currentProfile.p] =    GQLMixture.estimate_mixture(currentProfile.modelList(),
			    					        profileSet.ghmm_seqs, 10, 0.01, 					 			                        				        currentProfile.fixed_models)
	            currentProfile.cluster = GQLMixture.decode_mixture(currentProfile.p, num_clusters)
		elif(estimationMethod == 'C'):
		    [currentProfile.ml,
		     currentProfile.cluster,
		      currentProfile.p] = GQLMixture.estimate_clustering(currentProfile.modelList(), profileSet.ghmm_seqs,
                                                                                10, 0.01,currentProfile.fixed_models)
		elif(estimationMethod == 'K'):
		    # kmeans method ... the squared error is kept in the ml variable
		    [currentProfile.cluster, centroids,currentProfile.ml, nfound] = PyclusterInterface.kmeans(profileSet,num_clusters,npass=1)

                currentProfile.clusters = []
                for i in xrange(num_clusters):
                   currentProfile.clusters.append([])
                for i, c in enumerate(currentProfile.cluster):
                    currentProfile.clusters[c].append(i)
		profiles.append(currentProfile)
            except OverflowError:
	        print "repetition failed %i"%j
	return (profiles)

def resultsEvaluation(clusteringProfile,viterbiGroups):

        #do viterbi
        if( viterbiGroups != 0):
	    cluster = clusteringProfile.cluster
            no_clusters = 0
            for j in range(num_clusters):
              i = 0
	      cluster_aux = GQLCluster.ProfileSubSet(clusteringProfile.profileSet,clusteringProfile.clusters[j])
  	      vitDecomp = ViterbiDecomposition.ViterbiDecomposition(cluster_aux.getSequenceSet(), clusteringProfile.modelList()[j])
              grouping = vitDecomp.computeFixedNumberDecomposition(viterbiGroups)
  	      i = 0

  	      for group in grouping:
                  for element in group:
                      cluster[cluster_aux.ids[element]] = no_clusters + i
                  i += 1

	      no_clusters =+i

	    clusteringProfile.cluster = cluster

	(contigencyTable,rand,sen,spe,cr)= GQLValidation.computeExternalIndices(clusteringProfile.profileSet.seq_classes,clusteringProfile.cluster,
	                                                                        clusteringProfile.profileSet.classes_no,max(clusteringProfile.cluster)+1)
	print contigencyTable
	return (rand, sen, spe, contigencyTable)



if __name__ == '__main__':

        import sys

        fileName = ""
        num_clusters = 2
        nr_states = "[2]"
        viterbiGroups = 0
        repetitions = 30
        initializationMethod = 'R'
        estimationMethod = 'M'
        inputModel = ""
	labels = 0
	noOfClusters = 0



        argv = sys.argv

        print argv

        try:
            opts, args = getopt.getopt(argv[1:], 'h:c:e:i:l:m:r:s:v:no')
        except getopt.error, msg:
             raise Usage(msg)

        # process options
        for o, a in opts:
	    print o,a
            if o in ('-h'):
                print __doc__
                exit()
            if o in ('-c'):
                # Number of components
                num_clusters = int(a)
            if o in ('-s'):
                # Number of states
                int(a)
		nr_states = "["+a+"]"
            if o in ('-v'):
                # Nunber of groups in the viterbi decomposition
                viterbiGroups = int(a)
            if o in ('-r'):
                # Nunber of repetitions
                repetitions = int(a)
            if o in ('-m'):
                # Model file as input
                inputModel = a
            if o in ('-l'):
                # Model file as input
                labels = a
            if o in ('-e'):
                # Type of estimation
		if a in ('M','C','K'):
                  estimationMethod = a
		else:
		  raise Usage('Wrong option for -e')
            if o in ('-no'):
	        noOfClusters=1
            if o in ('-i'):
                # Type of estimation
		if a in ('R','M','K','L1','L2','L3'):
                  initializationMethod = a
		else:
		  raise Usage('Wrong option for -i')


	fileName = argv[-1]
        if fileName == "":
            raise Usage('No models file specified')

        if(initializationMethod == 'M'):
            repetitions = 1
	    if (inputModel == ""):
    	        raise Usage('option -i M requires the definition of a initial model')

        if(initializationMethod == 'L3'):
            repetitions = 1

	if(initializationMethod in ('L1','L2') ):
	    labels = int(a)/1000.0
	    if (labels == 0):
    	        raise Usage('option -i L requires the definition of the percentage of labels')


        profileSet = GQLCluster.ProfileSet()
	if "sqd" in fileName:
	    sequenceSet = ghmm.SequenceSet(ghmm.Float(),fileName)
            profileSet.ReadDataFromDSequences(sequenceSet,fileName)
	else:
            profileSet.ReadDataFromCaged(fileName)
        duration = len(profileSet[0])
        random = random_models_args()

	profiles = replications(profileSet,random,repetitions,initializationMethod,estimationMethod,inputModel,labels)


        # printing the results of all replications
        outfile = open("res-"+fileName+"-%i-%s-%s-%s-%i.txt"%(num_clusters,nr_states,initializationMethod,estimationMethod,viterbiGroups),'w')
        outfileCt = open("ctt-"+fileName+"-%i%s-%s-%s-%i.txt"%(num_clusters,nr_states,initializationMethod,estimationMethod,viterbiGroups),'w')

        rands = []
        sens = []
        spes = []
	aics = []
	bics = []
	ibics = []

        for i,p in enumerate(profiles):
            (rand, sen, spe, contigencyTable) = resultsEvaluation(p,viterbiGroups)

	    rands.append(rand)
	    sens.append(sen)
	    spes.append(spe)

    	    if(noOfClusters):
	        aic = GQLValidation.computeAIC(p.ml,p.modelList())
	    	bic = GQLValidation.computeBIC(p.ml,len(profileSet),p.modelList())
		ibic = GQLValidation.computeICL(p.ml,len(profileSet),p.modelList(),p.p)
	        outfile.write("\t%f\t%f\t%f"%(aic,bic,ibic))
		aics.append(aic)
		bics.append(bic)
		ibics.append(ibic)


	    outfile.write("%i\t%f\t%f\t%f\t%f\n"%(repetitions,rand, sen, spe,p.ml))

	    outfileCt.write(printTable(contigencyTable))

            for m in p.modelList():
	       m.write(fileName+"-%i-%s-%s-%s-%i-%i.smo"%(i,nr_states,initializationMethod,estimationMethod,viterbiGroups,i))

        outfile.close()
        outfileCt.close()

        # printing the results of all replications
        outfile = open("res-"+fileName+"-%s.txt"%(nr_states),'a')

	if( repetitions == 1):
	    rands.append(rands[0])
	    sens.append(sens[0])
	    spes.append(spes[0])

        if(initializationMethod in ('L1','L2','L') ):
 	    outfile.write("%f\t%f\t%f\t%f\t%f\t%f\t%f\t"%(labels,MLab.mean(rands),
							      MLab.std(rands),MLab.mean(spes),
							      MLab.std(spes),MLab.mean(sens),
							      MLab.std(sens)))
	    if( noOfClusters):
 	        outfile.write("%f\t%f\t%f\t%f\t%f\t%f\t%f\t"%(labels,MLab.mean(aics),
							      MLab.std(aics),MLab.mean(bics),
							      MLab.std(bics),MLab.mean(ibics),
							      MLab.std(ibics)))

	    outfile.write(" ".join(argv)+"\n")
	else:
 	    outfile.write("%i\t%f\t%f\t%f\t%f\t%f\t%f\t"%(num_clusters,MLab.mean(rands),
	    						      MLab.std(rands),MLab.mean(spes),
							      MLab.std(spes),MLab.mean(sens),
							      MLab.std(sens)))
	    if( noOfClusters):
 	        outfile.write("%f\t%f\t%f\t%f\t%f\t%f\t"%(MLab.mean(aics),
							      MLab.std(aics),MLab.mean(bics),
							      MLab.std(bics),MLab.mean(ibics),
							      MLab.std(ibics)))
	    outfile.write(" ".join(argv)+"\n")


	outfile.close()











